
#ifndef BANDEDALIGNER
#define BANDEDALIGNER

#include "CBT.hh"


#define mGapScore -2
#define mMatchScore 1
#define mMisMatchScore -1

// Score for things that fall outside of the band.  
#define OUTSIDE_BAND_SCORE -100000000

const int TOPFLAG= 1;
const int LEFTFLAG=2;
const int DIAGFLAG=4;

// Note, I want Banded Aligner to be efficient in the case of doing lots and lots of 
// banded alignments of read sized sequences.   For this purpose, I think it is useful to 
// pre-allocate the array to some maximum size and reuse it.  That is, BandedAligner will
// be optimized for the following kind of use:
//
// 
// BandedAligner BA;
// BA.SetMaxBandSize(MAXBANDSIZE);
// BA.SetMaxSeqLength(MAXSEQLENGTH);
// 
// for(int i = 0;i < ReadPairs.size();i++){
//   BA.setSource(S,ss,se);
//   BA.seqQuery(Q,qs,qe);
//   int score = BA.ComputeScore();
//   Sequence AlignS,AlignQ;
//   BA.GetAlignment(AlignS,AlignQ);
// }
// 
// Now, if you haven't set a Max size, BandedAligner should just set these values
// according to the sequence at hand.  
// 

class AlignmentTable : public vector<vector<int> >{
 public:


  void alloc(int band,int seqlength){
    resize(band);
    for(int i =0;i < band;i++){
      (*this)[i].resize(seqlength);
    }
    
    for(int i = 0;i < size();i++){
      for(int j = 0;j < (*this)[i].size();j++){
	(*this)[i][j] = 0;
      }
    }
  }

  void print(ostream &fout){
    for(int i = 0;i < size();i++){
      for(int j = 0;j < (*this)[i].size();j++){
	fout<<(*this)[i][j]<<"\t";
      }
      fout<<endl;
    }
  }

  void printvirtual(ostream &fout,int bsize){
    for(int r = 0;r < bsize;r++){
      for(int c = 0;c < bsize;c++){

      }
    }

  }



};


class BandedAligner{
 public:

  Sequence *pSource;
  Sequence *pQuery;

  int  QStart,QEnd;
  int  SStart, SEnd;
  int  BandSize;

  int  MaxBandSize;
  int  MaxSeqLength;

  AlignmentTable ScoreTable;
  AlignmentTable TraceBack;

  BandedAligner(Sequence &Source,int sstart,int send,Sequence &Query,int qstart,int qend){
    pSource = &Source;
    SStart = sstart;
    SEnd = send;
    pQuery = &Query;
    QStart = qstart;
    QEnd = qend;
    BandSize = -1;
    MaxSeqLength = -1;
  }

  BandedAligner(Sequence &Source,int sstart,int send,Sequence &Query,int qstart,int qend,int bs){
    pSource = &Source;
    SStart = sstart;
    SEnd = send;
    pQuery = &Query;
    QStart = qstart;
    QEnd = qend;
    BandSize = bs;
    MaxSeqLength = -1;
  }

  BandedAligner(){
    BandSize = -1;
    MaxSeqLength = -1;
  }


  void setMaxSeqLength(int maxSeqLen){
    MaxSeqLength = maxSeqLen;
    if (BandSize > 0){
      ScoreTable.alloc(2*BandSize+1,MaxSeqLength);
      TraceBack.alloc(2*BandSize+1,MaxSeqLength);
    }
  }

  void setBandSize(int bs){
    BandSize = bs;
    if (MaxSeqLength > 0){
      ScoreTable.alloc(2*BandSize+1,MaxSeqLength);
      TraceBack.alloc(2*BandSize+1,MaxSeqLength);
    }
  }

  void setSource(Sequence &Source,int sstart,int send){
    pSource = &Source;
    SStart = sstart;
    SEnd = send;
  }

  void setQuery(Sequence &Query,int qstart, int qend){
    pQuery = &Query;
    QStart = qstart;
    QEnd = qend;
  }

  int ComputeScore();  
  void GetAlignment(Sequence &AlignSource,Sequence &AlignQuery);


  void setFlag(int flag,int &cell){
    cell = cell | flag;
  }
  
  bool isSet(int flag,int cell){
    return(cell & flag);
  }

};



#endif 
