#ifndef INDEXEDACEFILE
#define INDEXEDACEFILE

char tempbuf[500];

class IndexedAceFile{
 public:
  AceFile A;
  ifstream mFin;
  
  String2Uint64Map Contig2Offset;
  
  void Open(string AceName){
    mFin.open(AceName.c_str());
  }
  
  int GetNumContigs(){
    return(Contig2Offset.size());
  }
  

  inline bool GetNextCOLine(istream &fin,string &line){
    bool bFeatureFound = false;
    do{
      fin.getline(readbuf,READ_BUFFER_SIZE);
      if (readbuf[0] == 'C'){
	if (readbuf[1] == 'O'){
	  line = readbuf;
	  return(true);
	}
      }
    }while(!fin.eof());

    return(false);
  }

  inline bool GetNextCOName(istream &fin,string &name){
    bool bFeatureFound = false;
    do{
      fin.getline(readbuf,READ_BUFFER_SIZE);
      if (readbuf[0] == 'C'){
	if (readbuf[1] == 'O'){
	  int tidx = 0;
	  // Find the next whitespace, and assign this subsequenct to name. 
	  for(int i = 3;i < READ_BUFFER_SIZE;i++){	    
	    if ((readbuf[i] == ' ') || (readbuf[i] == '\t')){
	      tempbuf[tidx] = '\0';
	      name = tempbuf;
	      return(true);
	    }else{
	      tempbuf[tidx++] = readbuf[i];
	    }
	  }	  
	}
      }
    }while(!fin.eof());

    return(false);
  }


  void ReadIndex(string IdxIn){
    ifstream fin;
    fin.open(IdxIn.c_str());
    do{
      string line;
      GetLine(fin,line);
      int offset = 0;
      string cName = SplitNext(line,offset);
      uint64 ctgoffset = NextUInt64(line,offset);
      Contig2Offset[cName] = ctgoffset;
    }while(!fin.eof());
    fin.close();
  }

  void WriteIndex(string IdxOut){
    ofstream fout;
    fout.open(IdxOut.c_str());
    String2Uint64Map::iterator it;
    for(it = Contig2Offset.begin();it != Contig2Offset.end();it++){
      fout<<(*it).first<<"\t"<<(*it).second<<endl;
    }
  }

  void CreateIndex(){
    uint64 ctgoffset;
    uint64 gcount;
    uint64 eoloffset;

    uint64 scale,lastscale = 0;

    do{
      //      string line;
      // if(GetNextLineContaining(mFin,"CO ",line)){
      //      if (GetNextCOLine(mFin,line)){
      string cName;
      if (GetNextCOName(mFin,cName)){
	// This is where it is now, at the end of the line....
	eoloffset = mFin.tellg();
	// Subtract last read to get the beginning of CO line
	gcount = mFin.gcount();
	ctgoffset = eoloffset - gcount;
	//int offset = 2;
	//	string cName = SplitNext(line,offset);
	//cout<<cName<<"\t"<<ctgoffset<<endl;
	Contig2Offset[cName] = ctgoffset;
      }    


      // Print a message every time the 10 millions digit changes. 
      // Processing time should be proprotional to number of bytes 
      // processed, not number of contigs, so this should produce somewhat
      // uniformly spaced (in time) feedback on indexing progress. 
      scale = ctgoffset / 50000000;
      if (scale != lastscale){
	lastscale = scale;
	cout<<Contig2Offset.size()<<"\t"<<cName<<"\t"<<ctgoffset<<endl;
      }

    }while(!mFin.eof());

    
    // GCC leaves the flags in a weird state when you reach eof(), so clear them. 
    mFin.clear();
    
  }
  
  void GetContigSequence(string CName,Sequence &Seq,SequenceQuality &Qual){
    A.resize(1);
    
    uint64 coffset = 0;
    if (Contig2Offset.Contains(CName)){
      coffset = Contig2Offset[CName];
    }else{
      cerr<<"FATAL ERROR: GetContigSequence: Can't find "<<CName<<endl;
      exit(1);
    }
   
    mFin.seekg(coffset,ios::beg);
    string line;
    int cNum = 0;
     
    if (FindTokenPlusWS(mFin,"CO",line)){
      A.ParseCO(mFin,line,cNum);    
      if (FindTokenPlusWS(mFin,"BQ",line)){
	line = A.ParseBQ(mFin,line,cNum);
      }
    
      A.CalculateContigTrims(A.mContigTrim);
      
      A[0].GetTrimmedDepadded(Seq,Qual);    
    }
    A.clear();
  }


  //  Extracts a contig into an existing acefile object. 
  void GetContig(string CName,AceFile &SubsetA){
    int cIdx = SubsetA.size();
    SubsetA.resize(SubsetA.size()+1);
    
    uint64 coffset = 0;
    if (Contig2Offset.Contains(CName)){
      coffset = Contig2Offset[CName];
    }else{
      cerr<<"FATAL ERROR: GetContigSequence: Can't find "<<CName<<endl;
      exit(1);
    }
    
    mFin.seekg(coffset,ios::beg);
    string line;
    if (FindTokenPlusWS(mFin,"CO",line)){
      SubsetA.ParseCO(mFin,line,cIdx);    
      if (FindTokenPlusWS(mFin,"BQ",line)){
	line = SubsetA.ParseBQ(mFin,line,cIdx);
	if (FindTokenPlusWS(mFin,"AF",line)){	 			
	  line= SubsetA.ParseAF(mFin,line,cIdx);
	  if (FindTokenPlusWS(mFin,"BS",line)){
	    line = SubsetA.ParseBS(mFin,line,cIdx);
	    if (FindTokenPlusWS(mFin,"RD",line)){
	      line = SubsetA.ParseRD(mFin,line,cIdx);
	    }
	  }
	}
      }
    }
  }






  void GetContig(string CName,Contig &C){
    A.resize(1);
    
    uint64 coffset = 0;
    if (Contig2Offset.Contains(CName)){
      coffset = Contig2Offset[CName];
    }else{
      cerr<<"FATAL ERROR: GetContigSequence: Can't find "<<CName<<endl;
      exit(1);
    }
   
    mFin.seekg(coffset,ios::beg);
    string line;
    int cNum = 0;
     
    if (FindTokenPlusWS(mFin,"CO",line)){
      A.ParseCO(mFin,line,cNum);    
      if (FindTokenPlusWS(mFin,"BQ",line)){
	line = A.ParseBQ(mFin,line,cNum);
      }
      
      C = A[0];
    }
    A.clear();
  }

  
  void Close(){
    mFin.close();
    // GCC does not clear these flags, which will cause bizarre behavior
    // if you try to reopen with the same stream object.  So clear them 
    // explicitly now to get the behavior we would expect. 
    mFin.clear();
  }
};

#endif 


 

