#include "Aligner.hh"


int cscore[][16] ={{   1  , 1 ,  1 ,  1 ,  1 , -1 ,  -1 ,  -1 , 1 , -1 ,  -1 , -1  , 1  , -1 , -1 , -1 },
                  {   -1 , 1 , -1 ,  -1 , 1 ,  1 ,  1 ,    1 , -1 , 1 ,  -1 , -1 , -1 ,  1 ,  -1 ,  -1 },
                  {   -1 ,-1 ,  1 , -1  , -1 , -1 , 1 ,  -1 ,  1 ,  1 ,  1  ,  1 , -1 , -1 ,  1  ,  -1},
                  {   -1 , -1 , -1 , 1  , -1 , -1 , -1 , 1  ,  -1 , -1 , -1 ,  1 ,  1 ,  1 ,  1  ,  1}};

//-------------------------------------------------------
// * getMatch
//-------------------------------------------------------
// 
void 
Aligner::getMatch(Subsequence &res1,Subsequence &res2){

  Traceback tb = mB0;
  int i = tb.i,j = tb.j;

  mEnd1 = i;
  mEnd2 = j;
  int padcount1 = 0,padcount2 = 0;

  //cout<<"i = "<<i<<" j = "<<j<<endl;
  tb = next(tb);
  while(tb.i >=0){
    //cout<<"next i= "<<tb.i<<" next j= "<<tb.j<<endl;

    if (i == tb.i) {
      res1+='-';
      padcount1++;
    }else {
      res1+=mSeq1[i-1];
    }
	 
    if (j == tb.j){
      res2+='-';
      padcount2++;
    }else {
      res2+=mSeq2[j-1];
    }

    //    cout<<res1<<endl;
    //cout<<res2<<endl;
    //tb.Print(cout);
    //cout<<endl<<endl;
	 
    i = tb.i; j = tb.j;
	 
    tb = next(tb);
  }

  mStart1 = i;
  mStart2 = j;
  mEnd1 = mEnd1 - padcount1;
  mEnd2 = mEnd2 - padcount2;

  res1.Reverse();
  res2.Reverse();
}

//  int mStart1, mStart2;
//  int mEnd1,mEnd2;

//-------------------------------------------------------
// * PrintTable
//-------------------------------------------------------
void PrintTable(Table<int> &T){
  for(int i = 0;i < T.vsize();i++){
    for(int j = 0;j < T.hsize();j++){
      cout<<T[i][j]<<" ";
    }
    cout<<endl;
  }
}


//-------------------------------------------------------
// * NWAligner
//-------------------------------------------------------
//
NWAligner::NWAligner(Substitution *psub, int GapCost,Subsequence &s1,Subsequence &s2)
  :BasicAligner(psub,GapCost,s1,s2)
{
  int n = mN,m = mM;
  Traceback tb;
  for(int i = 1;i <=n;i++){
    F[i][0] = -(mGapCost*i);
    B[i][0] = tb(i-1,0);
  }
  
  for(int j = 1;j <=m;j++){
    F[0][j] = -(mGapCost*j);
    B[0][j] = tb(0,j-1);
  }
  
  //  cout<<"NW n = "<<n<<" m="<<m<<endl;
  for(int i = 1;i <=n;i++){
    for(int j = 1;j <=m;j++){
      int s = (*mpSub)(s1[i-1],s2[j-1]);
      int val = max(F[i-1][j-1]+s,F[i-1][j]-mGapCost,F[i][j-1]-mGapCost);
      F[i][j] = val;
      //cout<<i<<","<<j<<" F[i][j]="<<val<<endl;
      if (val == F[i-1][j-1]+s)
	B[i][j] = tb(i-1,j-1);
      else if (val == F[i-1][j]-mGapCost)
	B[i][j] = tb(i-1,j);
      else if (val == F[i][j-1]-mGapCost)
	B[i][j] = tb(i,j-1);
      else cerr<<"Error "<<endl;

      //cout<<"B[i][j]="<<B[i][j].i<<","<<B[i][j].j<<endl;
    }
    mB0 = tb(n,m);
  }
  //  B.Print();
}



//-------------------------------------------------------
// * getMatch
//-------------------------------------------------------
//
void 
SWAligner3::getMatch(Subsequence &res1,Subsequence &res2){
  /*
  int n = mN,m = mM;
  for(int k = 0;k <=1;k++){
    for(int i = 1;i <=n;i++){
      for(int j = 1;j <=m;j++){
	if (B3[k][i][j].i == -1){
	  cout<<k<<"\t"<<i<<"\t"<<j<<"\t";
	  B3[k][i][j].Print(cout);
	  cout<<endl;
	}
      }
    }
  }
  */
  
  Traceback3 tb = mB0;
  int k = tb.k,i = tb.i,j = tb.j;

  tb = next(tb);
  while(tb.i >=0){

    cout<<k<<"\t"<<i<<"\t"<<j<<"\t";
    tb.Print(cout);
    cout<<endl<<endl;

    // The primary sequence reconstruction proceeds as normal...
    if (i == tb.i) res1+='-';
    else res1+=mSeq1[i-1];
	 
    // For the secondary sequence, need to deconvolve as we go. 
    if (j == tb.j) res2+='-';
    else {
      if (k == 0){
	res2+=mSeq2[j-1];
      }else{
	res2+=mSeq3[j-1];
      }
    }

    //   cout<<res1<<endl;
    //cout<<res2<<endl;
	 
    k = tb.k;i = tb.i; j = tb.j;
	 
    tb = next(tb);
  }
  res1.Reverse();
  res2.Reverse();
}


//-------------------------------------------------------
// * SWAligner3
//-------------------------------------------------------
// Special Smith Waterman for doing simultaneous alignment-deconvolution. 
//
SWAligner3::SWAligner3(Substitution *psub, int GapCost,Subsequence &s1,Subsequence &s2,
		      Subsequence &s3)
{
  mpSub = psub;
  mGapCost = GapCost;
  mSeq1 = s1;
  mSeq2 = s2;	 
  mSeq3 = s3;

  mN = s1.size();
  mM = s2.size();

  F3.resize(2);
  F3[0].Allocate(mN+1,mM+1);
  F3[1].Allocate(mN+1,mM+1);
  B3.resize(2);
  B3[0].Allocate(mN+1,mM+1);
  B3[1].Allocate(mN+1,mM+1);

  // Initially mark entire traceback table with stop symbol. 
  Traceback3 tbinit(-1,-1,-1);
  InitTraceback(tbinit);   

  int n = mN,m = mM;
  int maxi = n, maxj = m,maxk=0;
  int maxval = NegInf;

  Traceback3 tb;
  for(int k = 0;k <=1;k++){
    for(int i = 1;i <=n;i++){
      for(int j = 1;j <=m;j++){
	
	// There are two scores to consider, depending on which base 
	// you deconvolve to.  
	int s;
	if (k==0){
	  s = (*mpSub)(s1[i-1],s2[j-1]);
	}else{
	  s = (*mpSub)(s1[i-1],s3[j-1]);
	}

	//	int val = max(0,F3[i-1][j-1]+s,F3[i-1][j]-mGapCost,F3[i][j-1]-mGapCost);
	int val = max(0,
		      F3[0][i-1][j-1]+s,
		      F3[0][i-1][j]-mGapCost,
		      F3[0][i][j-1]-mGapCost,
		      F3[1][i-1][j-1]+s,
		      F3[1][i-1][j]-mGapCost,
		      F3[1][i][j-1]-mGapCost);

	F3[k][i][j] = val;
	
	if (val == 0){
	  B3[k][i][j] = tb(-1,-1,-1); // Terminator
	}else if (val == F3[0][i-1][j-1]+s){
	  B3[k][i][j] = tb(0,i-1,j-1);
	}else if (val == F3[1][i-1][j-1]+s){
	  B3[k][i][j] = tb(1,i-1,j-1);
	  cout<<k<<"\t"<<i<<"\t"<<j<<"\t";
	  B3[k][i][j].Print(cout);
	  cout<<endl;
	}else if (val == F3[0][i-1][j]-mGapCost){
	  B3[k][i][j] = tb(0,i-1,j);
	}else if (val == F3[1][i-1][j]-mGapCost){
	  B3[k][i][j] = tb(1,i-1,j);
	}else if (val == F3[0][i][j-1] - mGapCost){
	  B3[k][i][j] = tb(0,i,j-1);
	}else if (val == F3[1][i][j-1] - mGapCost){
	  B3[k][i][j] = tb(1,i,j-1);
	}else{
	  cerr<<"ERROR SW"<<endl;
	}
	
	if (val > maxval){
	  maxval = val;
	  maxk=k;maxi = i;maxj = j;
	}
      }
      mB0 = tb(maxk,maxi,maxj);
    }
  }
  //  B.Print();
}



//-------------------------------------------------------
// * SWAligner
//-------------------------------------------------------
//
//
SWAligner::SWAligner(Substitution *psub, int GapCost,Subsequence &s1,Subsequence &s2)
  :BasicAligner(psub,GapCost,s1,s2)
{

  int n = mN,m = mM;
  int maxi = n, maxj = m;
  int maxval = NegInf;

  Traceback tb;
  for(int i = 1;i <=n;i++){
    for(int j = 1;j <=m;j++){
      int s = (*mpSub)(s1[i-1],s2[j-1]);
      //      cout<<s1[i-1]<<"\t"<<s2[j-1]<<"\t"<<s<<endl;

      int val = max(0,F[i-1][j-1]+s,F[i-1][j]-mGapCost,F[i][j-1]-mGapCost);
      //      cout<<"val="<<val<<endl;
      F[i][j] = val;

      if (val == 0){
	B[i][j] = tb(-1,-1); // Terminator
      }else if (val == F[i-1][j-1]+s){
	B[i][j] = tb(i-1,j-1);
      }else if (val == F[i-1][j]-mGapCost){
	B[i][j] = tb(i-1,j);
      }else if (val == F[i][j-1] - mGapCost){
	B[i][j] = tb(i,j-1);
      }else{
	cerr<<"ERROR SW"<<endl;
      }

      if (val > maxval){
	maxval = val;
	maxi = i;maxj = j;
      }
    }
    mB0 = tb(maxi,maxj);
  }

  //B.Print();

}




//-------------------------------------------------------
// * NWLinearAligner
//-------------------------------------------------------
//
NWLinearAligner::NWLinearAligner(Substitution *psub, int GapCost,Subsequence &s1,Subsequence &s2)
  :LinearAligner(psub,GapCost,s1,s2){

  }

void
NWLinearAligner::ComputeScore(Subsequence &s1Seq,Subsequence &s2Seq)
{
  int n = mN, m = mM;
  register int u = n/2;
  C.Allocate(2,m+1);
  C.Init(0); // Make sure it's all zero to begin with. 

  int gapCost = mGapCost;
  register int i,j,s,val,Fojm,Foj,Fnjm,jm;

  register int newcol = 1, oldcol = 0;
  for(j = 0;j<=m;j++){
    F[newcol][j] = -(gapCost*j);
  }

  char *s1 = new char[s1Seq.size()];
  char *s2 = new char[s2Seq.size()];
  for(i = 0;i < s1Seq.size();i++) s1[i] = s1Seq[i];
  for(i = 0;i < s2Seq.size();i++) s2[i] = s2Seq[i];

  Substitution &S = *mpSub;

  for(i = 1;i<=n;i++){
    swap(newcol,oldcol);
    F[newcol][0] = -(gapCost * i);
    for(j = 1;j <=m;j++){
      jm = j-1;
      s = S(s1[i-1],s2[jm]);
      Fojm = F[oldcol][jm]+s;
      Foj = F[oldcol][j]-gapCost;
      Fnjm = F[newcol][jm]-gapCost;
      //      val = max(F[oldcol][j-1]+s,F[oldcol][j]-mGapCost,F[newcol][j-1]-mGapCost);
      val = max(Fojm,Foj,Fnjm);
      F[newcol][j] = val;
      if (i == u)
	C[newcol][j] = j;
      else{
	//	if (val == F[oldcol][j-1]+s){
	if (val == Fojm){
	  C[newcol][j] = C[oldcol][jm];
	  //}else if (val == F[oldcol][j] -mGapCost){
	}else if (val == Foj){
	  C[newcol][j] = C[oldcol][j];
	  //}else if (val == F[newcol][j-1] -mGapCost){
	}else if (val == Fnjm){
	  C[newcol][j] = C[newcol][jm];
	} else cerr<<"ERROR NWLinear"<<endl;
      }
    }
  }  

  // record these things
  mOldCol = oldcol;
  mNewCol = newcol;
  mU = u;

  delete [] s1;
  delete [] s2;
  
}

//-------------------------------------------------------
// * getMatch
//-------------------------------------------------------
//
void 
NWLinearAligner::getMatch(Subsequence &r1,Subsequence &r2){
  int v = getV();
  //cout<<"u="<<mU<<" v="<<v<<" mN="<<mN<<" mM="<<mM<<endl;
  if (mN > 1 && mM > 1){

    Subsequence a1(mSeq1,0,mU-1);
    Subsequence b1(mSeq2,0,v-1);
    Subsequence a2(mSeq1,mU,mSeq1.size()-1);
    Subsequence b2(mSeq2,v,mSeq2.size()-1);

    //cout<<">>"<<mSeq1<<endl;
    //cout<<">>"<<mSeq2<<endl;
    //cout<<a1<<endl;
    //cout<<b1<<endl;
    //cout<<a2<<endl;
    //cout<<b2<<endl;
        
    NWLinearAligner  al1(mpSub,mGapCost,a1,b1);
    al1.ComputeScore(a1,b1);

    NWLinearAligner  al2(mpSub,mGapCost,a2,b2);
    al2.ComputeScore(a2,b2);

    Subsequence match1,match2;
    al1.getMatch(match1,match2);

    Subsequence match3,match4;
    al2.getMatch(match3,match4);

    r1 = match1+match3;
    r2 = match2+match4;	
	 
    //cout<<"r1.size="<<r1.size()<<" r2.size="<<r2.size()<<endl;
		 	 
  }else{
    NWAligner al(mpSub,mGapCost,mSeq1,mSeq2);
    al.getMatch(r1,r2);
  }  
}

//-------------------------------------------------------
// * SWLinearAligner
//-------------------------------------------------------
//
SWLinearAligner::SWLinearAligner(Substitution *psub, int GapCost,Subsequence &s1,Subsequence &s2)
  :LinearAligner(psub,GapCost,s1,s2){
  }


void
SWLinearAligner::ComputeScore(Subsequence &s1Seq,Subsequence &s2Seq)
{    
  register int n = mN, m = mM;
  //start.Allocate(2,m+1);
  startI.Allocate(2,m+1);
  startJ.Allocate(2,m+1);
  
  //  Traceback tb;
  //start.Init(tb(-1,-1));
  startI.Init(-1);
  startJ.Init(-1);
  
  int MaxVal = NegInf;
  int start1,start2,end1,end2;
  register int newcol = 1, oldcol = 0;

  register int i,j,val,Foj,Fojm,Fnjm;

  for(j = 0;j<=m;j++){
    //    start[newcol][j] = tb(0,j);
    startI[newcol][j] = 0;
    startJ[newcol][j] = j;
  }

  // Make local copies of class member variables to avoid 
  // dereference inside loop.  

  char *s1 = new char[s1Seq.size()];
  char *s2 = new char[s2Seq.size()];
  for(i = 0;i < s1Seq.size();i++) s1[i] = s1Seq[i];
  for(i = 0;i < s2Seq.size();i++) s2[i] = s2Seq[i];

  Substitution &S = *mpSub;
  int gapCost = mGapCost;


  int s,jm;
  
  for(i = 1;i <=n;i++){
    swap(newcol,oldcol);	 
    //start[newcol][0] = tb(i,0);
    startI[newcol][0] = i;
    startJ[newcol][0] = 0;

    for(j = 1;j <=m; j++){
      jm = j-1;
      //int s = (*mpSub)(s1[i-1],s2[j-1]);
      //      s = S(s1[i-1],s2[jm]);
      s = S(s1[i-1],s2[jm]);

      Foj = F[oldcol][j];
      Fojm = F[oldcol][jm];
      Fnjm = F[newcol][jm];
     
      // int val = max(0,F[oldcol][j-1]+s,F[oldcol][j]-mGapCost,F[newcol][j-1]-mGapCost);
      val = max(0,Fojm+s,Foj-gapCost,Fnjm-gapCost);
      F[newcol][j] = val;

      if (val == 0){           // Best alignment starts (and ends) here
	//	start[newcol][j] = tb(i,j);
	startI[newcol][j] =i;
	startJ[newcol][j] =j;
	//      else if (val == F[oldcol][j-1]+s){
      }else if (val == Fojm+s){
	//start[newcol][j] = start[oldcol][j-1];
	startI[newcol][j] = startI[oldcol][jm];
	startJ[newcol][j] = startJ[oldcol][jm];
	//      }else if (val == F[oldcol][j]-mGapCost){
      }else if (val == Foj-gapCost){
	//start[newcol][j] = start[oldcol][j];
	startI[newcol][j] = startI[oldcol][j];
	startJ[newcol][j] = startJ[oldcol][j];
	//      }else if (val == F[newcol][j-1]-mGapCost){
      }else if (val == Fnjm-gapCost){
	//start[newcol][j] = start[newcol][j-1];
	startI[newcol][j] = startI[newcol][jm];
	startJ[newcol][j] = startJ[newcol][jm];
      }else cerr<<"SW Linear ERROR"<<endl;
				
      if (val > MaxVal){
	MaxVal = val;
	//	Traceback sij = start[newcol][j];
	//	mStart1 = sij.i; mStart2 = sij.j;
	//mEnd1 = i; mEnd2 = j;

	start1 = startI[newcol][j];
	start2 = startJ[newcol][j];
	end1 = i; end2 = j;
      }
    }
  }
  
  delete [] s1;
  delete [] s2;
  
  mMaxVal = MaxVal;
  mStart1 = start1;
  mStart2 = start2; 
  mEnd1 = end1;
  mEnd2 = end2; 
}

//-------------------------------------------------------
// * getMatch
//-------------------------------------------------------
//
void 
SWLinearAligner::getMatch(Subsequence &r1, Subsequence &r2)
{
  Subsequence sub1(mSeq1,mStart1,mEnd1-1);
  Subsequence sub2(mSeq2,mStart2,mEnd2-1);

  NWLinearAligner NW(mpSub,mGapCost,sub1,sub2);  
  NW.ComputeScore(sub1,sub2);
  NW.getMatch(r1,r2);  
}


















