#include "HitChainer.hh"

bool 
HitChainer::InWindow(Chain &C,SearchResult &H){
  int chainEndX = C.mSEnd;
  int chainEndY = C.mQEnd;

  int hitStartX = H.sourceStart();
  int hitStartY = H.queryStart();

  int x1 = chainEndX;
  int y1 = chainEndY;

  int x2 = hitStartX;
  int y2 = hitStartY;

  int bc =  y1-x1;
  int bp =  y1+x1;

  bool bL1 = y2 < x2 + bc + mWinWidth;
  bool bL2 = y2 > x2 + bc - mWinWidth;

  bool bL3 = y2 > -x2 + bp - mWinLength;
  bool bL4 = y2 < -x2 + bp + mWinLength;

  if (bL1 && bL2 && bL3 && bL4) return(true);
  else return(false);      
}

void 
HitChainer::ComputeChains(SearchResultVec &Hits,vector<Chain> &Chains)
{
  //  Hits.sortByQueryStart();
  Hits.sortBySourceStart();

  //  cout<<"=========== Sorted hits:"<<endl;
  //Hits.Print(cout);
  //cout<<endl;

  vector<bool> Used; 
  Used.resize(Hits.size());
  for(int i = 0;i < Used.size();i++) Used[i] = false;

  Chain tempChain;
  Segment tempSegment;
  tempSegment.mScore = 0;
  for(int i = 0;i < Hits.size();i++){  
    // If this hit hasn't already been used, try to start a chain with it...
    if (Used[i]) continue;
    
    tempSegment.mSBegin = Hits[i].sourceStart();
    tempSegment.mQBegin = Hits[i].queryStart();
    tempSegment.mSEnd = Hits[i].sourceEnd();//x1+merIncrement;
    tempSegment.mQEnd = Hits[i].queryEnd(); //y1 + merIncrement;
    tempSegment.mScore = 0;
    //tempSegment.mLength = Hits[i].sourceEnd() - Hits[i].sourceStart();
    
    tempChain.clear();
    tempChain.mSBegin = Hits[i].sourceStart();
    tempChain.mQBegin = Hits[i].queryStart();
    tempChain.mScore = 0;	
    tempChain.mTotalBases = 0;	
    //    tempChain.mSourceIdx = Hits[i].sourceIdx();
    //tempChain.mQueryIdx = Hits[i].queryIdx();
    tempChain.mSourceName = Hits[i].sourceName();
    tempChain.mSourceSize = Hits[i].sourceSize();
    tempChain.mQueryName = Hits[i].queryName();
    tempChain.mQuerySize = Hits[i].querySize();
    
    tempChain.mSEnd = tempSegment.mSEnd;
    tempChain.mQEnd = tempSegment.mQEnd;
    tempChain.mTotalBases+=Hits[i].sourceEnd() - Hits[i].sourceStart()+1;
    
    // Extend this chain as much as possible....
    // Since hits are sorted by source start, only need to examine hits after i. 
    // Moreover, can abort the loop whenever hit moves more than radius down the source axis...
    int radius = mWinWidth+mWinLength;
    for(int j = i+1;j < Hits.size();j++){

      if (Used[j]) continue;

      // Abort when too far down source to be a neighbor of end of chain...
      if (Hits[j].sourceStart() > (tempChain.mSEnd+radius)) break;

      // If this hit is within the window of the end of chain add it to the chain.
      if (InWindow(tempChain,Hits[j])){
	
	// Add the first sample if needed, since it was in limbo awaiting 
	// word on whether or not there would be a companion...
	if (tempChain.size() == 0) {
	  tempChain.push_back(tempSegment);
	  Used[i] = true;
	}
	  
	//	string sourceName = Hits[j].mSourceSeq.mName;
	tempSegment.mSBegin = Hits[j].mSourceStart;
	tempSegment.mQBegin = Hits[j].mQueryStart;
	tempSegment.mSEnd = Hits[j].mSourceEnd;//x1+merIncrement;
	tempSegment.mQEnd = Hits[j].mQueryEnd; //y1 + merIncrement;
	tempSegment.mScore = 0;
	
	tempChain.mSEnd = tempSegment.mSEnd;
	tempChain.mQEnd = tempSegment.mQEnd;
	tempChain.mTotalBases+=tempSegment.mSEnd - tempSegment.mSBegin+1;
	tempChain.push_back(tempSegment);
	Used[j] = true;
      }
    }
    Chains.push_back(tempChain);
  }
}



// This is less efficient than ComputeChains original since this may 
// visit a sample many many times.  In worst case, with no chains, 
// the complexity is n^2. 
void 
HitChainer::ComputeChains0(SearchResultVec &Hits,vector<Chain> &Chains)
{
  Hits.sortByQueryStart();
  Hits.sortBySourceStart();

  //  cout<<"=========== Sorted hits:"<<endl;
  //Hits.Print(cout);
  //cout<<endl;

  vector<bool> Used; 
  Used.resize(Hits.size());
  for(int i = 0;i < Used.size();i++) Used[i] = false;

  Chain tempChain;
  Segment tempSegment;
  tempSegment.mScore = 0;
  for(int i = 0;i < Hits.size();i++){  
    // If this hit hasn't already been used, try to start a chain with it...
    if (Used[i]) continue;

    tempSegment.mSBegin = Hits[i].sourceStart();
    tempSegment.mQBegin = Hits[i].queryStart();
    tempSegment.mSEnd = Hits[i].sourceEnd();//x1+merIncrement;
    tempSegment.mQEnd = Hits[i].queryEnd(); //y1 + merIncrement;
    tempSegment.mScore = 0;
    //tempSegment.mLength = Hits[i].sourceEnd() - Hits[i].sourceStart();
    
    tempChain.clear();
    tempChain.mSBegin = Hits[i].sourceStart();
    tempChain.mQBegin = Hits[i].queryStart();
    tempChain.mScore = 0;	
    tempChain.mTotalBases = 0;	
    //    tempChain.mSourceIdx = Hits[i].sourceIdx();
    //tempChain.mQueryIdx = Hits[i].queryIdx();
    tempChain.mSourceName = Hits[i].sourceName();
    tempChain.mSourceSize = Hits[i].sourceSize();
    tempChain.mQueryName = Hits[i].queryName();
    tempChain.mQuerySize = Hits[i].querySize();

    tempChain.mSEnd = tempSegment.mSEnd;
    tempChain.mQEnd = tempSegment.mQEnd;
    tempChain.mTotalBases+=Hits[i].sourceEnd() - Hits[i].sourceStart()+1;

    // Go through all the other hits trying to extend the chain...
    for(int j = 0;j < Hits.size();j++){

      // Skip self hits...
      if (i == j) continue;
      if (Used[j]) continue;

      // If this hit is within the window of the end of chain add it to the chain.
      if (InWindow(tempChain,Hits[j])){
	
	// Add the first sample if needed, since it was in limbo awaiting 
	// word on whether or not there would be a companion...
	if (tempChain.size() == 0) {
	  tempChain.push_back(tempSegment);
	  Used[i] = true;
	}
	  
	//	string sourceName = Hits[j].mSourceSeq.mName;
	tempSegment.mSBegin = Hits[j].mSourceStart;
	tempSegment.mQBegin = Hits[j].mQueryStart;
	tempSegment.mSEnd = Hits[j].mSourceEnd;//x1+merIncrement;
	tempSegment.mQEnd = Hits[j].mQueryEnd; //y1 + merIncrement;
	tempSegment.mScore = 0;
	
	tempChain.mSEnd = tempSegment.mSEnd;
	tempChain.mQEnd = tempSegment.mQEnd;
	tempChain.mTotalBases+=tempSegment.mSEnd - tempSegment.mSBegin+1;
	tempChain.push_back(tempSegment);
	Used[j] = true;
      }
    }
    Chains.push_back(tempChain);
  }
}

/*
//---------------------------------------------------------
// * WriteLV
//---------------------------------------------------------
//
void
BangEngine::WriteLV(ostream &fout,Sequence &Query,Chain &theChain)
{
  // This is the name of the particular sequence involved. 
  string QueryName = Query.mName;
  string SourceName = "";
  int oldSIdx = -10;

  long score = theChain.mScore;
		
  long sBegin = theChain.mSBegin+1;
  long sEnd = theChain.mSEnd+1;
  long chainSize = sEnd - sBegin +1;
		
  // require more than two samples...
  if (theChain.mTotalBases >= mMinScore){

    int sIdx = theChain.mSourceIdx;		
    // If the source has changed, output a new source/query description...
    if (oldSIdx != sIdx){
      oldSIdx = sIdx;
      SourceName = mMerLocMap.mSourceNames[sIdx];
      int sourceSize = mMerLocMap.mSourceSizes[sIdx];
			 
      // This is the name of the file....
      // theChainlearly redundant information, but blastz seems to output this so 
      // I've left it in for now.  I'll probably make this an option...
      fout<<"s {"<<endl;
      fout<<"  \""<<mSourceName<<"\" 1 "<<sourceSize<<" 0 1"<<endl;
      fout<<"  \""<<mQueryName<<"\" 1 "<<Query.size()<<" 0 1"<<endl;
      fout<<"}"<<endl;
			
      // This is the name of the contig within the file. 
      fout<<"h {"<<endl;
      fout<<"   \""<<SourceName<<"\""<<endl;
      fout<<"   \""<<QueryName<<"\""<<endl;
      fout<<"}"<<endl;			 
			
    }
		
    fout<<"a {"<<endl;
    fout<<"  s "<<score<<endl;
    // Must be one based, so +1;
    fout<<"  b "<<theChain.mSBegin+1<<" "<<theChain.mQBegin+1<<endl;
    fout<<"  e "<<theChain.mSEnd+1<<" "<<theChain.mQEnd+1<<endl;
    for(int j = 0;j < theChain.size();j++){
      long sstart = theChain[j].mSBegin;
      long send = theChain[j].mSEnd;
      long qstart = theChain[j].mQBegin;
      long qend = theChain[j].mQEnd;
      // +1 because output must be one based. 
      fout<<"  l "<<sstart+1<<" "<<qstart+1<<" "<<send+1<<" "<<qend+1<<" 100"<<endl;
    }
    fout<<"}"<<endl;
  }
}
*/

void 
HitChainer::WriteTabbedHeader(ostream &fout){

  fout<<"SourceName\tSourceSize\tQueryName\tQuerySize\tchainSize";
  fout<<"\tSourceBegin\tSourceEnd\tQueryBegin\tQueryEnd\tchainHits\tchainBases\tSense"<<endl;

}

//---------------------------------------------------------
// * ReadTabbed
//---------------------------------------------------------
// Reads the output from WriteTabbed.  Does not deal with 
// the segments of chains, only chains in aggregate. 
// 
void
HitChainer::ReadTabbed(istream &fin,vector<Chain> &Chains)
{
  string line;
  int offset = 0;
  Chain temp;
  
  // First line is a key to the settings used.  Just save it 
  // so it can be processed later if desired. 
  GetLine(fin,line);
  mSettingsKey = line;
  
  do{
    GetLine(fin,line);
    offset = 0;
    
    temp.mSourceName = SplitNext(line,offset);
    temp.mSourceSize = NextInt(line,offset);
    temp.mQueryName = SplitNext(line,offset);
    temp.mQuerySize = NextInt(line,offset);
    int chainSize = NextInt(line,offset);
    temp.mSBegin = NextInt(line,offset);
    temp.mSEnd = NextInt(line,offset);
    temp.mQBegin = NextInt(line,offset);
    temp.mQEnd = NextInt(line,offset);
    int chainHits = NextInt(line,offset);
    temp.mTotalBases = NextInt(line,offset);

    Chains.push_back(temp);
  }while(!fin.eof());
}

// Sname    SSize   QName                           QSize   ChainL  mSEnd   mSBegin  mQBegin  mQEnd   #mers   #bases
//chrX100k  100001  2kchrX1R0000001[26084,27084]    1001    998     27082   26085   -16      981      205     4100    -

//---------------------------------------------------------
// * WriteTabbed
//---------------------------------------------------------
// KJD: Sort of excessive and/or redundant to pass in Query. 
// 
// 
void
HitChainer::WriteTabbed(ostream &fout,Sequence &Query,Chain &theChain,bool bIsReverse)
{
  // This is the name of the particular sequence involved. 
  string QueryName = Query.mName;
  string SourceName = "";
		
  long sBegin = theChain.mSBegin+1;
  long sEnd = theChain.mSEnd+1;
  long chainLength = sEnd - sBegin +1;
  	       
  // Only output good hits...
  // if (theChain.mScore >= mMinScore){

  if (theChain.mTotalBases >= mMinScore){
    SourceName = theChain.mSourceName;
    int sourceSize = theChain.mSourceSize;
    
    fout<<SourceName<<"\t"<<sourceSize<<"\t"<<QueryName<<"\t"<<Query.size()<<"\t";
    fout<<"\t"<<chainLength<<"\t";

    if (bIsReverse) fout<<theChain.mSEnd+1<<"\t"<<theChain.mSBegin+1<<"\t";
    else fout<<theChain.mSBegin+1<<"\t"<<theChain.mSEnd+1<<"\t";

    fout<<theChain.mQBegin+1<<"\t"<<theChain.mQEnd+1<<"\t";
    fout<<theChain.size()<<"\t"<<theChain.mTotalBases;
    

    if (bIsReverse) fout<<"\t-"<<endl;
    else fout<<"\t+"<<endl;
    
    //    float avgCnt = (float)theChain.totalCount /(float)theChain.mScore;
    //fout<<avgCnt<<endl;
  }
}


