//========================================================================================
// SimpleMerHash.cc
// ----------------------------------------------------------------------------
// $Id: //CBT-depot/CBT++/src/SimpleMerHash.cc#7 $
// ------------------------------------------------------------------------------
// $Log: $
//
//========================================================================================
//.......1.........2.........3.........4.........5.........6.........7.........8.........9
//

#include "SimpleMerHash.hh"

//State mState;
//Word *mHash; // [HASHSIZE]; 
//Count *mCount; // [HASHSIZE] = {0}; 
//Index mNmers; // global nmer count 
//Index mDistinct;
//Index mReads; /* global read count */

//long mHashSize;

//int mProc;

char tempbuf2[5000];

//unsigned long  mCollisions;
//unsigned long long  mWalkSteps;
//unsigned long  mTotalInsertions;

// Former constants.  Now run-time variables. 
//int BASESUNUSED=0; // Default is 32-mer. 
//int BITSUNUSED=(BASESUNUSED*BITSPERBASE);

//--------------------------------------
// * ShiftIn
//--------------------------------------
// Don't call bases twice in same function call because it uses 
// a static char array.
// ShiftIn(nmer, reverseMer, Letter, ngood, proc);
void 
SimpleMerHash::ShiftIn(Word &w, Word &r,Word b,int &g,int &p) 
{
  w <<= (BITSPERBASE + BITSUNUSED);  
  w >>= BITSUNUSED; 
  w |= b;

  // KJD FIX This doesn't work with my encoding. Must fix all 
  // Reverse complement computations before switching to my encoding. 
  r >>= BITSPERBASE; 
  r |= (~b & BASEMASK) << (BITSPERWORD - (BITSPERBASE + BITSUNUSED)); 
  g++; 
  
  // Have we gotten a word's worth of bits yet?  If 
  // so, add the smaller of word/revcomp(word) to the 
  // hash.   
  if (g >= (BASESPERWORD-BASESUNUSED)) { 
    Word tw; 
    tw = (w <= r)? w : r; 
    AddCount(tw, p); 
  }
}

//--------------------------------------
// * Add
//--------------------------------------
// Adds mers from standard input stream. 
void
SimpleMerHash::Add(istream &fin)
{
  printf("Hash size is: %ld\n", mHashSize);
  Word nmer = 0;
  Word rev = 0;
  char pc, c;
  int ngood = 0;
  State s = FASTA;
  while ((c = fin.get()) != EOF) {
    if (s == DESCRIP) { 
      if (c == '\n') {
	s = FASTA;
	ngood = 0;
      }
    }else { // s == FASTA
      switch (c) {
	case '>': {
	  s = DESCRIP;
	  mReads++;
	  // A little something for the folks at home. 
	  if (! (mReads & 0xFFFF)) {
	    cerr<<"\r "<<mReads<<" reads\t"<<mNmers<<" nmers\t";
	    cerr<<mDistinct<<"distinct in hash ("<<((mDistinct*100UL)/mNmers)<<"%)"<<endl;
	  }
	  break;
	}
	case 'a': case 'A': {
	  ShiftIn(nmer, rev, Abits, ngood, mProc);
	  break;
	}
	case 'c': case 'C': {
	  ShiftIn(nmer, rev, Cbits, ngood, mProc);
	  break;
	}
	case 'g': case 'G': {
	  ShiftIn(nmer, rev, Gbits, ngood, mProc);
	  break;
	}
	case 't': case 'T': {
	  ShiftIn(nmer, rev, Tbits, ngood, mProc);
	  break;
	}
	case ' ': case '\t': case '\n': case '*': {
	  // Ignore
	  break;
	}
	default: {
	  nmer = 0;
	  rev = 0;
	  ngood = 0;
	}
      }
    }
  }
}


//--------------------------------------
// * CString2Mer
//--------------------------------------
// 
Word
SimpleMerHash::CString2Mer(const char * MerStr)
{
  Word nmer =0;
  int slen = strlen(MerStr);
  for(int i = 0;i< slen;i++){
    char c = MerStr[i];
    switch (c) {
      case 'a': case 'A': {
	ShiftInMer(nmer, Abits);
	break;
      }
      case 'c': case 'C': {
	ShiftInMer(nmer, Cbits);
	break;
      }
      case 'g': case 'G': {
	ShiftInMer(nmer, Gbits);
	break;
      }
      case 't': case 'T': {
	ShiftInMer(nmer,Tbits);
	break;
      }
      case ' ': case '\t': case '\n': case '*': {
	// Ignore
	break;
      }
      default: {
	nmer = 0;
      }
    }
  }
  return(nmer);
}


//--------------------------------------
// * String2Mer
//--------------------------------------
// 
Word
SimpleMerHash::String2Mer(string &MerStr)
{
  Word nmer;
  int slen = MerStr.size();
  for(int i = 0;i< slen;i++){
    char c = MerStr[i];
    switch (c) {
      case 'a': case 'A': {
	ShiftInMer(nmer, Abits);
	break;
      }
      case 'c': case 'C': {
	ShiftInMer(nmer, Cbits);
	break;
      }
      case 'g': case 'G': {
	ShiftInMer(nmer, Gbits);
	break;
      }
      case 't': case 'T': {
	ShiftInMer(nmer,Tbits);
	break;
      }
      case ' ': case '\t': case '\n': case '*': {
	// Ignore
	break;
      }
      default: {
	nmer = 0;
      }
    }
  }
  return(nmer);
}

//--------------------------------------
// * Mer2CString
//--------------------------------------
// 
char *
SimpleMerHash::Mer2CString(Word w) {
  //  static char b[BASESPERWORD+1];
  //	char *b = new char(BASESPERWORD+1);

  char base = '\0';
  int basebits;
  int i;
  unsigned long long bugNum = (BASESPERWORD - BASESUNUSED);

  // This looks buggy because the last value assigned to "base" is not
  // used.  That is an extra assignment with no effect; all the bases
  // and null are properly stored in the (1 + BASESPERWORD - BASESUNUSED) iterations.
  for (i = (BASESPERWORD - BASESUNUSED); i >= 0; i--) {
    tempbuf2[i] = base;
    basebits = w & BASEMASK;
    w >>= BITSPERBASE;
    switch (basebits) {
      case Abits: 
	base = 'a';
	break;
      case Cbits: 
	base = 'c';
	break;
      case Gbits: 
	base = 'g';
	break;
      case Tbits: 
	base = 't';
	break;
    }
  }
  return tempbuf2;
}

//--------------------------------------
// * ThinWord
//--------------------------------------
// This was replicated many places.  It's 
// a hit to make it a function call, but this ensures 
// the implementation is consistent.  Can make it a 
// macro if the performance is noticable. 
//
// Returns true if word should be thinned, false otherwise. 
// 
bool 
SimpleMerHash::ThinWord(Word w){
	Index low32,high32;
	low32 = w & 0xFFFFFFFF;
	high32 = w >> 32;
	if ((low32 ^ high32) % hashThinningPrime != hashSlice){
		return(true);
	}else{
		return(false);
	}
}

//--------------------------------------
// * AddCount
//--------------------------------------
//
void 
SimpleMerHash::AddCount(Word w, int proc) {
  Index probe, i;
  Count t;
  Word temp;
  Count bitcount = 0;
  
  //#ifndef UNIBATCH
  //for (temp = lo ^ up; temp; temp >>= 5) {
  //	 bitcount ^= temp & 0x1F;
  // }
  //if (bitcount != proc) {
  //	 return;
  // }
  //#endif

  mNmers++;

	if (ThinWord(w)) return;

  findslot(probe,w);
  // fprintf(stderr, "%s %ld\n", Mer2CString(w), probe);
  if (! mCount[probe]) {
    mDistinct++;
    mCount[probe] = 1;
    mHash[probe] = w;
  }else if (mHash[probe] != w) {
    fprintf(stderr, "\nFull hash on word %s, probe %ld, insertion %ld\n", 
	    Mer2CString(w), probe, mDistinct); 
    exit(1);
  }else if (((Count)(t = mCount[probe] + 1)) > 0) {
    mCount[probe] = t;
  }
}

//--------------------------------------
// * ShiftInMer
//--------------------------------------
//
void 
SimpleMerHash::ShiftInMer(Word &w, Word b) 
{
  w <<= (BITSPERBASE + BITSUNUSED);  
  w >>= BITSUNUSED; 
  w |= b;
  //cout<<"ShiftInMer :"<<w<<endl;
}

//char basebits = (MerStr[i] & 0x06) >> 1;
//shiftinmer(w,basebits);


//--------------------------------------
// * Add
//--------------------------------------
// 
void 
SimpleMerHash::Add(string &MerStr,int count){
  //Word w = String2Mer(MerStr);  // Replace with inline to try performance. 

  Word w;
  int slen = MerStr.size();
  for(int i = 0;i< slen;i++){
    char c = MerStr[i];
    switch (c) {
      case 'a': case 'A': {
	shiftinmer(w, Abits);
	break;
      }
      case 'c': case 'C': {
	shiftinmer(w, Cbits);
	break;
      }
      case 'g': case 'G': {
	shiftinmer(w, Gbits);
	break;
      }
      case 't': case 'T': {
	shiftinmer(w,Tbits);
	break;
      }
      default: {
	cerr<<"ILLEGAL CHARACTER: "<<c<<" in mer attempted add to hash."<<endl; 
	return;
      }
    }
  }
  Add(w,count);
}

//--------------------------------------
// * Add
//--------------------------------------
// 
void 
SimpleMerHash::Add(char* MerStr,int count){
  //Word w = String2Mer(MerStr);  // Replace with inline to try performance. 

  Word w;
  //int slen = MerStr.size();
  int stop = 32 - BASESUNUSED;
  for(int i = 0; i < stop; i++){
    char c = MerStr[i];
    switch (c) {
      case 'a': case 'A': {
	shiftinmer(w, Abits);
	break;
      }
      case 'c': case 'C': {
	shiftinmer(w, Cbits);
	break;
      }
      case 'g': case 'G': {
	shiftinmer(w, Gbits);
	break;
      }
      case 't': case 'T': {
	shiftinmer(w,Tbits);
	break;
      }
      default: {
	cerr<<"ILLEGAL CHARACTER: "<<c<<" in mer attempted add to hash."<<endl; 
	return;
      }
    }
  }
  // cout<<"Adding " << MerStr << "; w = " << Mer2CString(w) << ", count = " << count <<endl;
  bool bThinned = Add(w,count);
  if (bThinned){
    cout<<"Thinned: "<<MerStr<<endl;
  }
}

Word 
SimpleMerHash::CreateWord(const char* MerStr){
  Word w;
  //int slen = MerStr.size();
  int stop = 32 - BASESUNUSED;
  for(int i = 0; i < stop; i++){
    char c = MerStr[i];
    switch (c) {
      case 'a': case 'A': {
	shiftinmer(w, Abits);
	break;
      }
      case 'c': case 'C': {
	shiftinmer(w, Cbits);
	break;
      }
      case 'g': case 'G': {
	shiftinmer(w, Gbits);
	break;
      }
      case 't': case 'T': {
	shiftinmer(w,Tbits);
	break;
      }
      default: {
	cerr<<"ILLEGAL CHARACTER: "<<c<<" in mer attempted add to hash."<<endl; 
	return(0);
      }
    }
  }
  return(w);
}


//--------------------------------------
// * Add
//--------------------------------------
// 
bool 
SimpleMerHash::Add(Word w,int count){

  mTotalInsertions++;

  Index probe,i;
  Count t;
  Word temp;
  Count bitcount = 0;
  const Count maxCount = (Count) 0xFFFF;
  
  if (ThinWord(w)) {
    return(true); // KJD temporary debugging test. 
  }
  findslot(probe,w);

  // fprintf(stderr, "%s %ld\n", Mer2CString(w), probe);

  if (mCount[probe] && mHash[probe] != w) {
    fprintf(stderr, "\nFull hash on word %s, probe %ld, insertion %ld\n", 
	    Mer2CString(w), probe, mDistinct); 
    exit(1);
  }
  else {
    // Not already in the hash. or in hash with different count
    mHash[probe] = w;
    mCount[probe] = (count > maxCount)? maxCount : count;
  }
  return(false);
}

//--------------------------------------
// * Mer2Word
//--------------------------------------
// Mer is in the wrong encoding, so need to translate it to 
// a Word and call Contains on that.  This is just a stop gap 
// way to do this, in order to get something working quickly. 
// Efficiency will be addressed by changing hash encoding. 
//
// KJD Shoot me.  I can't believe how I encoded this. 
// acgtacgtacgt ggggccccttttaaaa
// 0123....
// 
// So given an input sequence or string, the numbering will
// go from left to right as above.  The leftmost base will be zero. 
// My mer encoding puts bases into the upper or lower 32 bits
// depending on whether the base index i < 16 or not.  If < 16, 
// it goes into the Lower32, if > 16, it goes into the upper. 
// Worse, I shift the bases in from right to left.  The result, 
// the bits of the Mer correspond to:
// 
// aaaattttccccgggg tgcatgcatgcatgca
// 
// Horrid.   I don't know what I was thinking, but for now
// it will be easier to have this one function here handle the
// weirdness than it will be to change all of the places in 
// MerHashMap where the encoding is crazy.   At least I think it
// will be....
//
Word
SimpleMerHash::Mer2Word(Mer M)
{
  const unsigned char transtable = // 0xb4; // 10110100 (11 -> 10, 10 -> 11, etc)
    (Abits << (Akjd * 2)) | (Cbits << (Ckjd * 2)) | (Gbits << (Gkjd * 2)) | (Tbits << (Tkjd * 2));
  // printf("%x\n", transtable);
  Index upper32,lower32;
  upper32 = M.mUpper32;
  lower32 = M.mLower32;

  // cout << "Upper32: " << hex << upper32 << endl;
  // cout << "Lower32: " << lower32 << dec << endl;
  Word w = 0;

  Index phhup = 0;
  Index phhlo = 0;
  Index revup = 0;
  Index revlo = 0;
  char kjdbits, phhbits;

  // OK, this is crazy.  My Lower32 bits will become
  // the upper 32 bits for the hash, and as I shift my 
  // bits out to the right, they are added to the left. 

  // Shift in the lower 32 bits (2 bits at a time). 
  // Set up size for divisions
#define BITSPERINDEX (sizeof(Index) * 8)

  for(int i = 0;i< 16;i++){
    kjdbits = (lower32 & 0x00000003) << 1; // grab the lower two bits, * 2
    lower32 = lower32>>2; // Shift them out. 

    phhbits = (transtable >> kjdbits) & 0x3;
    phhup <<= BITSPERBASE;
    phhup |= phhbits;

    revlo >>= BITSPERBASE;
    revlo |= ((Index) phhbits) << (BITSPERINDEX - BITSPERBASE);

    // Shift in the upper 32 bits (2 bits at a time). 

    kjdbits = (upper32 & 0x00000003) << 1; // grab the lower two bits, * 2. 
    upper32 = upper32>>2; // Shift them out. 
	 
    phhbits = (transtable >> kjdbits) & 0x3;
    phhlo <<= BITSPERBASE;
    phhlo |= phhbits;

    revup >>= BITSPERBASE;
    revup |= ((Index) phhbits) << (BITSPERINDEX - BITSPERBASE);
  }
  w = (((Word) phhup) << 32) | phhlo;
  Word r = ~((((Word) revup) << 32) | revlo);
  if (BASESUNUSED) {
    w >>= BITSUNUSED;
    r <<= BITSUNUSED;
    r >>= BITSUNUSED;
  }
  // cout<<"w= "<< hex << w << " " << Mer2CString(w) <<endl;
  // cout<<"r= "<< r << " " << Mer2CString(r) <<endl;
  w = (w <= r)? w : r;
  // cout<<"min= "<< w << dec << " " << Mer2CString(w) <<endl;
  return(w);
}

//--------------------------------------
// * Contains
//--------------------------------------
// More efficient would be to build M and rcM at the 
// same time, and pass both in.  
// 
// 
bool
SimpleMerHash::Contains(Mer M)
{
  Word w = Mer2Word(M);	// now returns canonical Mer (lesser of fwd, rev)
  if (ContainsFwd(w)) return(true);
  else return(false);
  //    else{
  //  	 w = RevComp(w);
  //  	 if (ContainsFwd(w)) return(true);
  //  	 else return(false);
  //    }
} 

//--------------------------------------
// * Contains
//--------------------------------------
//
bool
SimpleMerHash::Contains(Mer M,int &count){
  Word w = Mer2Word(M);	// now returns canonical Mer (lesser of fwd, rev)
  Count ccount = 0;
  if (ContainsFwd(w,ccount)){
    count = (int) ccount;
    return(true);
  }
  else {
    count = 0;
    return(false);
  }
  //    }else{
  //  	 w = RevComp(w);
  //  	 if (ContainsFwd(w,ccount)){
  //  		count = (int) ccount;
  //  		return(true);
  //  	 }else{
  //  		count = 0;
  //  		return(false);
  //  	 }
  //    }
}

//--------------------------------------
// * ContainsFwd
//--------------------------------------
//
bool
SimpleMerHash::ContainsFwd(const char *MerStr)
{
  Word Mer = CString2Mer(MerStr);
  return(ContainsFwd(Mer));
}

//--------------------------------------
// * Contains
//--------------------------------------
//
bool
SimpleMerHash::Contains(Word w)
{
  Index probe,i;
  Count t;
  Word temp;
  Count bitcount = 0;
      
	if (ThinWord(w)) return(false);

  findslot(probe,w);

  if ((mHash[probe] == w) && (mCount[probe] > 0)) return(true);
  else return(false);  
}


//--------------------------------------
// * ContainsFwd
//--------------------------------------
//
bool
SimpleMerHash::ContainsFwd(Word w)
{
  Index probe,i;
  Count t;
  Word temp;
  Count bitcount = 0;
      
	if (ThinWord(w)) return(false);

  findslot(probe,w);

  if ((mHash[probe] == w) && (mCount[probe] > 0)) return(true);
  else return(false);  
}


//--------------------------------------
// * Contains
//--------------------------------------
//
bool
SimpleMerHash::ContainsFwd(Word w,Count &count)
{
  Index probe,i, low32, high32;
  Count t;
  Word temp;
  Count bitcount = 0;

	if (ThinWord(w)) return(false);

  findslot(probe,w);

  if ((mHash[probe] == w) && (mCount[probe] > 0)){
    count = mCount[probe];
    return(true);
  }else {
    count = 0;
    return(false); 
  } 
}
//--------------------------------------
// * RevComp
//--------------------------------------
//
Word 
SimpleMerHash::RevComp(Word w) {
  Index i;
  Word rev = 0;
  
  // The word w is treated as a string of 2-bit bases
  // (A == 01, T == 11, C == 01, G == 10)
  // So long as we reverse the string basewise,
  // we just need to bit-complement to get the final answer.  
  for (i = 1; i <= BASESPERWORD; i++) {
    //  Shift result left and add in last base of input; then advance input
    rev = (rev << BITSPERBASE) + (w & BASEMASK);
    w >>= BITSPERBASE;
  }
  return ~rev; // bitwise complement == basewise complement 
}

void 
SimpleMerHash::SetMerSize(int merSize){
  BASESUNUSED = 32 - merSize; 
  BITSUNUSED=(BASESUNUSED*BITSPERBASE);

  mMerSize = merSize;
}


//-------------------------------------------
// * ReadMerDist
//-------------------------------------------
//
void 
SimpleMerHash::ReadMerDist(istream &fin,int minCount)
{
  long merCount = 1,addCount = 0;
  do{
    fin.getline(readbuf,READ_BUFFER_SIZE);
    // Find the first whitespace
    int i;
    for(i = 0;i< READ_BUFFER_SIZE;i++){		
      if ((readbuf[i] == '\t') | (readbuf[i] == ' ')) break;
    }

    // Convert remainder of the line to an int. 
    int count = atoi(readbuf+i+1);		
    if (count >= minCount){		 
      Add(readbuf,count);
      addCount++;
    }
    if (!(merCount%1000000)){
      cout<<"Mers seen:"<<merCount<<endl;
    }
    merCount++;

  }while(!fin.eof());
}

//-------------------------------------------
// * ReadRepeatMers
//-------------------------------------------
//
void 
SimpleMerHash::ReadMerDist(string MerFileName,int minCount)
{
  izipstream zin;
  zin.open(MerFileName.c_str());
  ReadMerDist(zin,minCount);
  zin.close();
}


//-------------------------------------------
// * GuessHashSizeFromDist
//-------------------------------------------
// It's inconvenient to always have to manually 
// guess the hash size.  In the case of a kmer dist table 
// we know the size, it's simply 'wc distFile'.  However, 
// this can take a very long time to run, and as a separate
// step is annoying.... so just get the file size and 
// come up with a conservative estimate of the size. 
// 
// 
long 
SimpleMerHash::GuessHashSizeFromDist(string FileName)
{
  ifstream fin;
  fin.open(FileName.c_str());
  long l = fin.tellg();
  fin.seekg(0,ios::end);
  long m = fin.tellg();

  long s = m-l;

  // Now, is this a gzipped file?  If so, need to 
  // inflate this size considerably 
  int idx = FileName.find(".gz");
  if (idx>=0) {
    s = s * 5;
  }

  // OK, so each entry in a distribution file will be a k-mer
  // a tab, and a number.   We want to overestimate the size 
  // slightly, so to be conservative let's say the numbers are 
  // only a single digit.  
  long linesize = mMerSize + 1 + 1;
	
  long numMers = (long)((double)s / (double)linesize);

  // Inflate the estimate by, say, 20%. 
  long estimate = (long)((double)numMers * (double)1.2);

  return(estimate);
}

