
#include "BangEngine.hh"

//.......1.........2.........3.........4.........5.........6.........7.........8.........9

//---------------------------------------------------------
// * AddSource
//---------------------------------------------------------
//
void 
BangEngine::AddSource(string SourceName){

  ifstream fin;
  fin.open(SourceName.c_str());  
  //mMerLocMap.Add(fin,mRepeatModel);
  mMerLocMap.Add(fin);
  fin.close();

  if (mbDebug){
    Mer2SequenceLocationMap<MerHashF,BigSequenceLocation>::iterator it;
    cout<<"ENTIRE HASH: "<<endl;
    for(it = mMerLocMap.begin();it != mMerLocMap.end();it++){
      cout<<(*it).first.ToString()<<endl;
      for(int i = 0;i < (*it).second.size();i++){
	BigSequenceLocation &SL = (*it).second[i];
	cout<<"\t"<<SL.sourceStart()<<endl;
      }
    }
    cout<<endl<<endl;
  }

}

//---------------------------------------------------------
// * SetupRepeatModel
//---------------------------------------------------------
//
void 
BangEngine::SetupRepeatModel(string DistFile, int maxMerCount){
  int minMerCount = 1;
  mRepeatModel.ReadDist(DistFile,minMerCount);
  mRepeatModel.setMaxCount(maxMerCount);
}
 
//---------------------------------------------------------
// * Compare
//---------------------------------------------------------
//
void 
BangEngine::GetHits(Sequence &Query,
		    SearchResultVec &FwdHits,
		    SearchResultVec &RevHits)
{  
  int sampleDelta = 1;
  int QSize = Query.size();
  //  int merstop = QSize - Mer::mMerSize +1;
  int merstop = QSize;
  
  // FwdHits are added to end, so only want to update 'new' hits. 
  // fwdHitIdx points to the first 'new' hit.  Ditto revHitIdx. 
  int fwdHitIdx = 0,revHitIdx=0;

  //  cout<<"BangEngine:: Query.mName="<<Query.mName<<endl;

  Mer fwMer;

  // NOTE: Given how ShiftInClean works, qIdx is the next base to be added to a mer, not the start of the mer. 
  for(int qIdx = 0;qIdx< merstop;qIdx++){

    fwMer.ShiftInClean(Query,qIdx);
    
    if (mbDebug) cout<<"qIdx="<<qIdx<<"\tMer="<<fwMer.ToString()<<endl;

    Mer rcMer(fwMer);
    rcMer.Reverse();

    // Only search thinned mers below the repeat cutoff.... 
    // Note:  thinning is based on smaller of fwd/rev mer. 
    // To save time and compute rev only once, this version takes
    // rev as argument. 
    if(!fwMer.isThinned(rcMer)){      
      if (mRepeatModel.isAcceptableFrequency(fwMer)){
	mMerLocMap.FindMerLocations(fwMer,FwdHits);	

	// FindMerLocations doesn't know anything about where the 
	// mers came from, so need to update the hits with this info. 
	for(int h = fwdHitIdx;h < FwdHits.size();h++){
	  SearchResult &H = FwdHits[h];
	  H.setQueryStart(qIdx - Mer::mMerSize +1);
	  H.setQueryEnd(qIdx);
	  //	  H.setQueryStart(qIdx);
	  //H.setQueryEnd(qIdx+Mer::mMerSize-1);
	  H.setQueryName(Query.mName);
	  H.setQuerySize(Query.size());
	  // save mer frequency somehow?	
	  fwdHitIdx = FwdHits.size();
	}      

	if (mbDebug){
	  cout<<"FWD: "<<fwMer.ToString()<<endl;
	  FwdHits.Print(cout);
	}

	mMerLocMap.FindMerLocations(rcMer,RevHits);

	//cout<<"BANG Engine: revHits.size="<<RevHits.size()<<endl;	
	// RevHits are expressed in Query coordinates.  We want hits to be expressed 
	// in ReverseQuery coordinates.  See diagrams below for illustration. 
	// 
	//           1         2 
	// 0123456789012345678901234567
	// ACTGATACATTGGGATACCCATGCCCAT  Query size=28
	//     ATACAT                    fwMer Query[4,9]
	//     ATGTAT                    rcMer Reverse(Query[4,9])
	
	//           1         2 
	// 0123456789012345678901234567
	// ATGGGCATGGGTATCCCAATGTATCAGT  ReverseQuery
	//                   ATGTAT      rcMer ReverseQuery[18,23]
	
	// rstart = Qsize - fsend -1;   rend = Qsize - fsstart -1;
	// rstart = 28 - 9 - 1 = 18    rend = 28 - 4 - 1 = 23 
	for(int h = revHitIdx;h < RevHits.size();h++){
	  SearchResult &H = RevHits[h];
	  int rawqstart = qIdx - Mer::mMerSize +1;
	  H.setQueryEnd(QSize - rawqstart -1);
	  H.setQueryStart(H.queryEnd() - Mer::mMerSize+1);

	  //	  H.setQueryEnd(QSize - qIdx -1);
	  //H.setQueryStart(H.queryEnd() - Mer::mMerSize+1);
	  H.setQueryName(Query.mName);
	  H.setQuerySize(Query.size());
	  // save mer frequency somehow?
	  revHitIdx = RevHits.size();
	}

	if (mbDebug){
	  cout<<"REV: "<<rcMer.ToString()<<endl;
	  RevHits.Print(cout);
	  cout<<endl;
	}
      }
    }
  }
}

//---------------------------------------------------------
// * Compare
//---------------------------------------------------------
// Compares a single query to the source returning chains. 
// 
void
BangEngine::Compare(Sequence &Query,vector<Chain> &FwdChains,vector<Chain> &RevChains){
  SearchResultVec FwdHits,RevHits; 
  
  GetHits(Query,FwdHits,RevHits);

  if (mbDebug){
    cout<<endl<<endl;

    cout<<"\nFwdHits:"<<endl;
    FwdHits.Print(cout);

    cout<<"RevHits:"<<endl;
    RevHits.Print(cout);
  }
    
  HitChainer H;
  H.mMinScore = mMinScore;

  //  cerr<<"BUG VALUE: fwd rev hits size: \t"<<FwdHits.size()<<"\t"<<RevHits.size()<<endl;

  H.ComputeChains(FwdHits,FwdChains);  
  H.ComputeChains(RevHits,RevChains);  
}

/*

void
BangEngine::FilterChains(vector<Chain> &FChains,vector<Chain> &RChains){

This might do several things.   Including:

* Finding only best hit, or best k hits, for a given query/source pair. 

* Examining segments to compute actual fraction of match in match range. 
  That is, if we have 1000 bases and 800 mer matches, what fraction of the 
  1000 bases is covered by the 800 mers?  

* Banded alignments of top candidates?  

}

*/

/*
void
BangEngine::ReadSourceForAlignments(string SourceName){
  FAFile FA;
  FA.Read(SourceName,mOptionalSource);
  for(int i = 0;i < mOptionalSource.size();i++){
    mSourceName2Idx[mOptionalSource[i].mName] = i;
  }
}

void 
BangEngine::WriteAlignments(ostream &fout,Sequence &Query,vector<Chain> &Chains){
  BandedAligner BA;
  BA.setBandSize(10);
  BA.setMaxSeqLength(2000);
  for(int i = 0;i < Chains.size();i++){
    int sourceIdx = mSourceName2Idx[Chains[i].mSourceName];
    Sequence &S = mOptionalSource[sourceIdx];
    BA.setSource(S,Chains[i].mSBegin,Chains[i].mSEnd);
    BA.setQuery(Query,Chains[i].mQBegin,Chains[i].mQEnd);
    int score = BA.ComputeScore();
    Sequence AlignS,AlignQ;
    BA.GetAlignment(AlignS,AlignQ);
    HitChainer H;
    H.WriteTabbed(fout,Query,Chains[i]);
    fout<<AlignS<<endl;
    fout<<AlignQ<<endl;    
  }
}
*/

//---------------------------------------------------------
// * Compare
//---------------------------------------------------------
// Does a comparison on a batch of query sequences, outputting chains as it goes. 
// Chains are either output in LAV format or as tabbed output.  
//
void 
BangEngine::Compare(istream &fin,ostream &fout)
{
  FASTAFile FA;
  
  char buffer[5000];
  buffer[0] = '1';
  
  int totalFwdHits = 0,totalRevHits=0;
  
  // Write out LAV header or tabbed field description. 
  HitChainer H;
  H.mMinScore = mMinScore; // KJD Init this in more than one place, awkward and error prone.
  if (!mbTabbedOut) H.WriteLVHeader(fout);
  else H.WriteTabbedHeader(fout);

  do{
    // This is not the fastest read function.... major place 
    // for optimization... probably the Sequence3Bit code is faster...    
    Sequence Query;	
    if (!FA.GetNextSequence(fin,buffer,5000,Query)) break;
	 
    // Can't do anything with a tiny query. 
    if (Query.size() < mMinQuerySize) break;
    
    vector<Chain> FwdChains,RevChains;
    Compare(Query,FwdChains,RevChains);

    //    cout<<"fwdchains.size()="<<FwdChains.size()<<endl;
    //cout<<"revchains.size()="<<RevChains.size()<<endl;

    //    FilterChains(FwdChains,RevChains);

    if (mbAlignments){
      //WriteAlignments(fout,Query,FwdChains);
      //WriteAlignments(fout,Query,RevChains);
    }else{
      //if (mbTabbedOut){
      bool bIsReverse = false;
      H.WriteTabbed(fout,Query,FwdChains,bIsReverse);
      bIsReverse = true;
      H.WriteTabbed(fout,Query,RevChains,bIsReverse);
      //}else{
      //H.WriteLV(fout,Query,FwdChains);
      //H.WriteLV(fout,Query,RevChains);
      //}
    }
  }while(!fin.eof());
  
  if (!mbTabbedOut) H.WriteLVFooter(fout);
}
