#include "BlastzEmulator.hh"

//---------------------------------------------------------
// * AddSource
//---------------------------------------------------------
//
void 
BlastzEmulator::AddSource(string SourceName){
  mSourceName = SourceName;
  ifstream fin;
  fin.open(SourceName.c_str());
  mSE.AddToHashMapSampled(fin,0,99999,mSourceSkip);

	cout<<"mSE.size=\t"<<mSE.size()<<"\t"<<endl;
	cout<<"mSE.mSourceNames.size=\t"<<mSE.mSourceNames.size()<<"\t"<<endl;

  fin.close();
}
 
//---------------------------------------------------------
// * AddSource
//---------------------------------------------------------
//
void 
BlastzEmulator::AddSourceThinned(string SourceName){
  mSourceName = SourceName;
  ifstream fin;
  fin.open(SourceName.c_str());
  
  mSE.AddToHashMapThinned(fin);
  //mSE.AddToHashMap(fin,0,1);
  //	mSE.AddToHashMapThinned(fin,0,500000,f,slice,mSMH);
  //mSE.AddToHashMapThinned2(fin,f,slice);
	
	cout<<"mSE.size=\t"<<mSE.size()<<"\t"<<endl;
	cout<<"mSE.mSourceNames.size=\t"<<mSE.mSourceNames.size()<<"\t"<<endl;
	
  fin.close();
}

//---------------------------------------------------------
// * ComputeRevChains
//---------------------------------------------------------
// Each hit will be extendended into a segment.  A segment is 
// a perfect match stretch of the sequence.   In the current implementation, 
// each chain will be made up of exactly one segment.
// In the future I may string segments together into chains which 
// are not perfect matches, and respect the distinction that blastz makes
// with that, but for now a chain will be the same as a segment. 
//
void 
BlastzEmulator::ComputeRevChains(Sequence &Query,MerSearchResultVec<BigMerLocation> &Hits,
																 ofstream &fout)
{
	

  // The coordinate system is likely to be very confusing.  See 
  // note at bottom of file. 

  Hits.SortByRevDiagonal();
	
  vector<bool> Used; 
  Used.resize(Hits.size());
  for(int i = 0;i < Used.size();i++) Used[i] = false;
  
  // Go through the hits building segments of perfect matches...
  int merIncrement = Mer::mMerSize -1;
  for(int h = 0; h < Hits.size();h++){
	 
    if (Used[h]){
      continue;  // Skip samples that have already been used. 
    }
	 
    // Start chain with this sample....
    Segment tempSegment;
    int x1 = Hits[h].SourceBaseIdx();
    int y1 = Hits[h].mQueryBaseIdx;

    int sIdx = Hits[h].SourceRead();

    tempSegment.mSBegin = x1;
    tempSegment.mQBegin = y1+merIncrement;
    tempSegment.mSEnd = x1+merIncrement;  // Sample 
    tempSegment.mQEnd = y1+merIncrement; 
    tempSegment.mScore = 1;
    Chain tempChain;
    tempChain.clear();
    tempChain.mSBegin = x1;
    tempChain.mQBegin = y1+merIncrement;
    tempChain.mScore = 0;
    tempChain.mSourceIdx = Hits[h].SourceRead(); // KJD Note: 19 bits for this. 

    mX0 = x1;
    mY0 = y1;

    //	 cout<<x1<<"\t"<<y1<<"\t";
    //cout<<tempSegment.mSBegin<<"\t"<<tempSegment.mSEnd<<"\t";
    //cout<<tempSegment.mQBegin<<"\t"<<tempSegment.mQEnd<<endl;

		int totalCount = 0;

    // Scan until we find a non-matching sample or we get to the next sequence...
    int merIncrement = Mer::mMerSize - 1;
    for(int h2 = h+1;h2 < Hits.size();h2++){	  	 
      int x2 = Hits[h2].SourceBaseIdx();
      int y2 = Hits[h2].mQueryBaseIdx;

			totalCount+=Hits[h2].mMerGlobalFrequency;

      int newSIdx = Hits[h2].SourceRead();
      if (newSIdx != sIdx) break;

      // Look for a sample within the next window.
      if (SampleInRevWindow(x1,y1,x2,y2,mWinWidth,mWinLength)){
	Used[h2] = true;
	tempSegment.mSEnd = x2+merIncrement;
	tempSegment.mQEnd = y2;	
	tempSegment.mScore++;
	
	//		 cout<<x1<<"\t"<<y1<<"  -> "<<x2<<"\t"<<y2<<"\t";
	//cout<<tempSegment.mSBegin<<"\t"<<tempSegment.mSEnd<<"\t";
	//cout<<tempSegment.mQBegin<<"\t"<<tempSegment.mQEnd<<endl;
	
	x1 = x2;
	y1 = y2;
      }else{ 
	break;
      }
    }
    //		cout<<"totalCount="<<totalCount<<"\tscore="<<tempSegment.mScore<<endl;
    tempChain.mAvgCount = (float)totalCount / (float)tempSegment.mScore;
    tempChain.totalCount = totalCount;
    
    //	 cout<<"FINAL:\t";
    // cout<<tempSegment.mSBegin<<"\t"<<tempSegment.mSEnd<<"\t";
    //cout<<tempSegment.mQBegin<<"\t"<<tempSegment.mQEnd<<endl;
    
    tempChain.mSEnd = tempSegment.mSEnd;
    tempChain.mQEnd = tempSegment.mQEnd;
    tempChain.mScore+=tempSegment.mScore;
    tempChain.push_back(tempSegment);
    
    if (mbTabbedOut){
      WriteTabbed(fout,Query,tempChain);
    }else{
      WriteLV(fout,Query,tempChain);
    }
  }
}


//---------------------------------------------------------
// * ComputeChains
//---------------------------------------------------------
// Each hit will be extendended into a segment.  A segment is 
// a perfect match stretch of the sequence.   In the current implementation, 
//
void 
BlastzEmulator::ComputeChains(Sequence &Query,MerSearchResultVec<BigMerLocation> &Hits,
															ofstream &fout)
{
  Hits.SortByDiagonal();

	//	cout<<endl;
	//cout<<"ComputeChains"<<endl;
	//Hits.Print();

  vector<bool> Used; 
  Used.resize(Hits.size());
  for(int i = 0;i < Used.size();i++) Used[i] = false;
  
  // Go through the hits building segments of perfect matches...
  int merIncrement = Mer::mMerSize -1;
  for(int h = 0; h < Hits.size();h++){
	 
    if (Used[h]){
      continue;  // Skip samples that have already been used. 
    }
	 
    // Start chain with this sample....
    Segment tempSegment;
    int x1 = Hits[h].SourceBaseIdx();
    int y1 = Hits[h].mQueryBaseIdx;

    int sIdx = Hits[h].SourceRead();

    tempSegment.mSBegin = x1;
    tempSegment.mQBegin = y1;
    tempSegment.mSEnd = x1+merIncrement;
    tempSegment.mQEnd = y1 + merIncrement;
    tempSegment.mScore = 1;
    Chain tempChain;
    tempChain.clear();
    tempChain.mSBegin = x1;
    tempChain.mQBegin = y1;
    tempChain.mScore = 0;
    tempChain.mSourceIdx = Hits[h].SourceRead(); // KJD Note: 19 bits for this. 

    mX0 = x1;
    mY0 = y1;

		int totalCount = 0;

    // Scan until we find a non-matching sample or we get to the next sequence...
    for(int h2 = h+1;h2 < Hits.size();h2++){	  	 
      int x2 = Hits[h2].SourceBaseIdx();
      int y2 = Hits[h2].mQueryBaseIdx;

			totalCount+=Hits[h2].mMerGlobalFrequency;

      int newSIdx = Hits[h2].SourceRead();
      if (newSIdx != sIdx) break;

      // Look for a sample within the next window.
      if (SampleInWindow(x1,y1,x2,y2,mWinWidth,mWinLength)){
				//cout<<"sample in window."<<endl;
				Used[h2] = true;
				tempSegment.mSEnd = x2+merIncrement;
				tempSegment.mQEnd = y2+merIncrement;	
				tempSegment.mScore++;

				//			 cout<<x1<<"\t"<<y1<<"  -> "<<x2<<"\t"<<y2<<"\t";
				//			 cout<<tempSegment.mSBegin<<"\t"<<tempSegment.mSEnd<<"\t";
				//			 cout<<tempSegment.mQBegin<<"\t"<<tempSegment.mQEnd<<endl;

				x1 = x2;
				y1 = y2;
      }else{ 
				//cout<<"sample NOT in window."<<endl;
				break;
      }
    } 

		tempChain.mAvgCount = (float)totalCount / (float)tempSegment.mScore;
		tempChain.totalCount = totalCount;
    tempChain.mSEnd = tempSegment.mSEnd;
    tempChain.mQEnd = tempSegment.mQEnd;
    tempChain.mScore+=tempSegment.mScore;		
    tempChain.push_back(tempSegment);
		
    if (mbTabbedOut){
			//cout<<"tempChain.size="<<tempChain.size()<<"\tscore="<<tempChain.mScore<<endl;
      WriteTabbed(fout,Query,tempChain);
    }else{
      WriteLV(fout,Query,tempChain);
    }
  }
}

//---------------------------------------------------------
// * SampleInRevWindow
//---------------------------------------------------------
// Define a window that the sample should be in.  Basically this 
// is a box 2*WinWidth wide and WinLength long centered on the 
// exact match diagonal line through the X0,Y0 sample (the first 
// sample in a run).   This region is defined by the region bounded
// by four lines. 
//
// Here the window is bounded by two parallel lines to the exact match line, 
// and two perpendicular lines, one going through x1,y1 and another going 
// through (x1+winLength,y1+winLength).   The region bounded by these four
// lines is the window. 
// 
inline bool
BlastzEmulator::SampleInRevWindow(int x1,int y1, int x2, int y2, int WinWidth, int WinLength)
{
  // y-intercept of the exact match line...
  // This is a fixed reference defined by the first sample in a series. 
  int b1 = mY0+mX0;

  // The two bounding parallel lines are:
  // Strictly speaking, the parallel line that is WinWidth (radius) across 
  // is y = -x + b1 +2a  where 2a = (1/sqrt(2))*WinWidth which is roughly 0.707*WinWidth
  // So, for this, I just take that as being exactly WinWidth, since the boundaries
  // are only heuristics anyway.  
  bool bLine4 = (y2 < (-x2 +b1 + WinWidth));
  bool bLine3 = (y2 > (-x2 +b1 - WinWidth));

  // y-intercept of the perpendicular line to the exact match line. 
  int b2 = y1-x1;
  bool bLine1 = (y2 < (x2 + b2));

  // y-intercept of the perpendicular line that is offset WinLength down from
  // the exact match line. 
  int b3 = y1-x1-2*WinLength;
  bool bLine2 = (y2 > (x2 + b3));

  if (bLine3 && bLine4 && bLine1 && bLine2){
    return(true);
  }else {
    return(false);
  }
}

//---------------------------------------------------------
// * SampleInWindow
//---------------------------------------------------------
// Define a window that the sample should be in.  
// Here the window is bounded by two parallel lines to the exact match line, 
// and two perpendicular lines, one going through x1,y1 and another going 
// through (x1+winLength,y1+winLength).   The region bounded by these four
// lines is the window. 
// 
inline bool
BlastzEmulator::SampleInWindow(int x1,int y1, int x2, int y2, int WinWidth, int WinLength)
{
  // y-intercept of the exact match line...
  // This is a fixed reference defined by the first sample in a series. 
  int b1 = mY0 - mX0; 

  // The two bounding parallel lines are:
  bool bLine3 = (y2 > (x2 +b1 - WinWidth));
  bool bLine4 = (y2 < (x2 +b1 + WinWidth));

  // y-intercept of the perpendicular line to the exact match line. 
  int b2 = y1+x1;
  bool bLine1 = (y2 > (-x2 + b2));

  // y-intercept of the perpendicular line that is offset WinLength from 
  // the exact match line. 
  int b3 = y1+x1+2*WinLength;
  bool bLine2 = (y2 < (-x2 + b3));

  if (bLine3 && bLine4 && bLine1 && bLine2){
    return(true);
  }else {
    return(false);
  }
}

//---------------------------------------------------------
// * Compare
//---------------------------------------------------------
//
void 
BlastzEmulator::Compare(istream &fin,ofstream &fout)
{
  FASTAFile FA;
  
  char buffer[5000];
  buffer[0] = '1';
  
  int totalFwdHits = 0,totalRevHits=0;
  
  // Write out LAV header or tabbed field description. 
  if (!mbTabbedOut) fout<<"#:lav"<<endl;
  else WriteTabbedHeader(fout);


  do{
    Sequence Query;
    MerSearchResultVec<BigMerLocation> FwdHits;
    MerSearchResultVec<BigMerLocation> RevHits; 
 
    vector<Chain>  FwdChains,RevChains;
	 
    // This is not the fastest read function.... major place 
    // for optimization... probably the Sequence3Bit code is faster...
    if (!FA.GetNextRead(fin,buffer,5000,Query)) break;
	 
    // Can't do anything with a tiny query. 
    if (Query.size() < 33) break;

    Compare(Query,FwdHits,RevHits);	 

    ComputeChains(Query,FwdHits,fout);
    ComputeRevChains(Query,RevHits,fout);
	 
    totalFwdHits+=FwdHits.size();
    totalRevHits+=RevHits.size();
	 	 
  }while(!fin.eof());
  
  if (!mbTabbedOut) fout<<"#:eof"<<endl;
  
	//  cout<<"totalFwdHits="<<totalFwdHits<<endl;
  //cout<<"totalRevHits="<<totalRevHits<<endl;
}


//---------------------------------------------------------
// * Compare
//---------------------------------------------------------
//
void 
BlastzEmulator::CompareThinned(istream &fin,ofstream &fout)
{
  FASTAFile FA;
  
  char buffer[5000];
  buffer[0] = '1';
  
  int totalFwdHits = 0,totalRevHits=0;
  
  // Write out LAV header or tabbed field description. 
  if (!mbTabbedOut) fout<<"#:lav"<<endl;
  else WriteTabbedHeader(fout);


  do{
    Sequence Query;
    MerSearchResultVec<BigMerLocation> FwdHits;
    MerSearchResultVec<BigMerLocation> RevHits; 
 
    vector<Chain>  FwdChains,RevChains;
	 
    // This is not the fastest read function.... major place 
    // for optimization... probably the Sequence3Bit code is faster...
    if (!FA.GetNextRead(fin,buffer,5000,Query)) break;
	 
    // Can't do anything with a tiny query. 
    if (Query.size() < 33) break;

    CompareThinned(Query,FwdHits,RevHits);	


		///		FwdHits.Print(); 
 

    ComputeChains(Query,FwdHits,fout);
    ComputeRevChains(Query,RevHits,fout);
	 
    totalFwdHits+=FwdHits.size();
    totalRevHits+=RevHits.size();
	 	 
  }while(!fin.eof());
  
  if (!mbTabbedOut) fout<<"#:eof"<<endl;
  
  cout<<"totalFwdHits="<<totalFwdHits<<endl;
  cout<<"totalRevHits="<<totalRevHits<<endl;
}




//---------------------------------------------------------
// * Compare
//---------------------------------------------------------
//
void 
BlastzEmulator::Compare(Sequence &Query,
												MerSearchResultVec<BigMerLocation> &FwdHits,
												MerSearchResultVec<BigMerLocation> &RevHits)
{
  
  int sampleDelta = 1;


  // KJD There seems to be some bug in FindMerLocations.  The 
  // last sample, sampleIdx=1418, is not returning any hits. 
  // I think I have noticed the same problem with atlas-overlapper, 
  //MerSearchResultVec<BigMerLocation> bugHits;
  //mSE.FindMerLocations(Query,1418,bugHits);
  //	cout<<"bugHits.size()="<<bugHits.size()<<endl;
  
  int merSize = Mer::mMerSize;
  for(int i = 0;i< Query.size()-merSize;i+=sampleDelta){		
    mSE.FindMerLocations(Query,i,FwdHits);
    mSE.FindRevMerLocations(Query,i,RevHits);
  }
  
 
  int rCount = 0;
  // KJD I forget... why are there redundant entries??
  // Now remove the redundant entries. 
  if (FwdHits.size()>1){
    for(int i = 1;i< FwdHits.size();i++){
      MerSearchResult<BigMerLocation> lastResult = FwdHits[i-1];
      if ((lastResult.mData == FwdHits[i].mData) && 
					(lastResult.mData1 == FwdHits[i].mData1)){
				FwdHits.erase(FwdHits.begin()+i);
				rCount++;
				i--; // Adjust i 
      }
    }
  }
  
  // Now remove the redundant reverse entries. 
  if (RevHits.size()>1){
    for(int i = 1;i< RevHits.size();i++){
      MerSearchResult<BigMerLocation> lastResult = RevHits[i-1];
      if ((lastResult.mData == RevHits[i].mData) && 
					(lastResult.mData1 == RevHits[i].mData1)){
				RevHits.erase(RevHits.begin()+i);
				rCount++;
				i--; // Adjust i since 
      }
    }
  }

  //	cout<<"rCount="<<rCount<<endl;

}



//---------------------------------------------------------
// * Compare
//---------------------------------------------------------
//
void 
BlastzEmulator::CompareThinned(Sequence &Query,
															 MerSearchResultVec<BigMerLocation> &FwdHits,
															 MerSearchResultVec<BigMerLocation> &RevHits)
{
  
  int sampleDelta = 1;


  // KJD There seems to be some bug in FindMerLocations.  The 
  // last sample, sampleIdx=1418, is not returning any hits. 
  // I think I have noticed the same problem with atlas-overlapper, 
  //MerSearchResultVec<BigMerLocation> bugHits;
  //mSE.FindMerLocations(Query,1418,bugHits);
  //	cout<<"bugHits.size()="<<bugHits.size()<<endl;
  
  int merSize = Mer::mMerSize;
  for(int i = 0;i< Query.size()-merSize;i++){		
		
    // Only search thinned mers below the repeat cutoff.... 
    Mer tempMer(Query,i);
		if(!tempMer.Thinned(f,slice)){
			// SMH only contains mers that are over the cutoff. 
			//			if (!mSMH.Contains(tempMer)){
			// Move this operation inside FindMerLocations so can record count also. 
			mSE.FindMerLocations(Query,i,FwdHits,mSMH);
			mSE.FindRevMerLocations(Query,i,RevHits,mSMH);
			//	}
		}
  }
  
  int rCount = 0;
  // KJD I forget... why are there redundant entries??
  // Now remove the redundant entries. 
  if (FwdHits.size()>1){
    for(int i = 1;i< FwdHits.size();i++){
      MerSearchResult<BigMerLocation> lastResult = FwdHits[i-1];
      if ((lastResult.mData == FwdHits[i].mData) && 
					(lastResult.mData1 == FwdHits[i].mData1)){
				FwdHits.erase(FwdHits.begin()+i);
				rCount++;
				i--; // Adjust i 
      }
    }
  }
  
  // Now remove the redundant reverse entries. 
  if (RevHits.size()>1){
    for(int i = 1;i< RevHits.size();i++){
      MerSearchResult<BigMerLocation> lastResult = RevHits[i-1];
      if ((lastResult.mData == RevHits[i].mData) && 
					(lastResult.mData1 == RevHits[i].mData1)){
				RevHits.erase(RevHits.begin()+i);
				rCount++;
				i--; // Adjust i since 
      }
    }
  }

  //	cout<<"rCount="<<rCount<<endl;

}



//---------------------------------------------------------
// * WriteLV
//---------------------------------------------------------
//
void
BlastzEmulator::WriteLV(ostream &fout,Sequence &Query,Chain &theChain)
{
  // This is the name of the particular sequence involved. 
  string QueryName = Query.mName;
  string SourceName = "";
  int oldSIdx = -10;

  long score = theChain.mScore;
		
  long sBegin = theChain.mSBegin+1;
  long sEnd = theChain.mSEnd+1;
  long chainSize = sEnd - sBegin +1;
		
  // require more than two samples...
  if (chainSize >= mMinScore){


    int sIdx = theChain.mSourceIdx;		
    // If the source has changed, output a new source/query description...
    if (oldSIdx != sIdx){
      oldSIdx = sIdx;
      SourceName = mSE.mSourceNames[sIdx];
      int sourceSize = mSE.mSourceSizes[sIdx];
			 
      // This is the name of the file....
      // theChainlearly redundant information, but blastz seems to output this so 
      // I've left it in for now.  I'll probably make this an option...
      fout<<"s {"<<endl;
      fout<<"  \""<<mSourceName<<"\" 1 "<<sourceSize<<" 0 1"<<endl;
      fout<<"  \""<<mQueryName<<"\" 1 "<<Query.size()<<" 0 1"<<endl;
      fout<<"}"<<endl;
			
      // This is the name of the contig within the file. 
      fout<<"h {"<<endl;
      fout<<"   \""<<SourceName<<"\""<<endl;
      fout<<"   \""<<QueryName<<"\""<<endl;
      fout<<"}"<<endl;			 
			
    }
		
    fout<<"a {"<<endl;
    fout<<"  s "<<score<<endl;
    // Must be one based, so +1;
    fout<<"  b "<<theChain.mSBegin+1<<" "<<theChain.mQBegin+1<<endl;
    fout<<"  e "<<theChain.mSEnd+1<<" "<<theChain.mQEnd+1<<endl;
    for(int j = 0;j < theChain.size();j++){
      long sstart = theChain[j].mSBegin;
      long send = theChain[j].mSEnd;
      long qstart = theChain[j].mQBegin;
      long qend = theChain[j].mQEnd;
      // +1 because output must be one based. 
      fout<<"  l "<<sstart+1<<" "<<qstart+1<<" "<<send+1<<" "<<qend+1<<" 100"<<endl;
    }
    fout<<"}"<<endl;
  }
}


void 
BlastzEmulator::WriteTabbedHeader(ostream &fout){

  fout<<"SourceName\tSourceSize\tQueryName\tQuerySize\tchainSize";
  fout<<"\tSourceBegin\tSourceEnd\tQueryBegin\tQueryEnd\tchainHits\tglobalmerSum\tavgcnt"<<endl;
}


//---------------------------------------------------------
// * WriteTabbed
//---------------------------------------------------------
//
void
BlastzEmulator::WriteTabbed(ostream &fout,Sequence &Query,Chain &theChain)
{
  // This is the name of the particular sequence involved. 
  string QueryName = Query.mName;
  string SourceName = "";

  long score = theChain.mScore;
		
  long sBegin = theChain.mSBegin+1;
  long sEnd = theChain.mSEnd+1;
  long chainSize = sEnd - sBegin +1;
		
	//	cout<<"minScore="<<mMinScore<<"\tchainSize="<<chainSize<<endl;

  // Only output good hits...
  if (chainSize >= mMinScore){
    int sIdx = theChain.mSourceIdx;		
    SourceName = mSE.mSourceNames[sIdx];
    int sourceSize = mSE.mSourceSizes[sIdx];
			 
    fout<<SourceName<<"\t"<<sourceSize<<"\t"<<QueryName<<"\t"<<Query.size()<<"\t";
    fout<<"\t"<<chainSize<<"\t";
    fout<<theChain.mSBegin+1<<"\t"<<theChain.mSEnd+1<<"\t";
    fout<<theChain.mQBegin+1<<"\t"<<theChain.mQEnd+1<<"\t";
		fout<<theChain.mScore<<"\t"<<theChain.totalCount<<"\t";
		float avgCnt = (float)theChain.totalCount /(float)theChain.mScore;
		fout<<avgCnt<<endl;
  }
}




//---------------------------------
// Note on sample coordinates. 
//
// The coordinates for FwdHits is as you would expect. 
// The coordinates for RevHits are not what one would 
// expect.   


// Here is how it works:
//
// The query sequence Q is sampled: [0,31], [32,63], ...
// Each sample [0,31] has it's reverse complement computed. 
// This RC-sample is compared to the source. 
// The coordinate returned for the query hit of a sample
// from [0,31] is 0.  If we ask, "Where is this in the 
// RC of the entire query sequence?", then the answer is 
// that it is at RCQuery.size()-1, which is the coordinate
// most people seem to expect for this operation. 
// 
// I may change this coordinate in the future.  
//

